from pytest_django_queries.utils import assert_type, raise_error


class Entry(object):
    BASE_FIELDS = [("test_name", "Test Name")]
    REQUIRED_FIELDS = [("query-count", "Queries")]
    OPTIONAL_FIELDS = [("duplicates", "Duplicated")]
    FIELDS = BASE_FIELDS + REQUIRED_FIELDS + OPTIONAL_FIELDS

    def __init__(self, test_name, module_name, data):
        """
        :param data: The test entry's data.
        :type data: dict
        """

        assert_type(data, dict)

        self._raw_data = data
        self.test_name = test_name
        self.module_name = module_name

        for field, _ in self.REQUIRED_FIELDS:
            setattr(self, field, self._get_required_key(field))

        for field, _ in self.OPTIONAL_FIELDS:
            setattr(self, field, self._get_key(field, "UNK"))

    def __getitem__(self, item):
        return getattr(self, item)

    @property
    def query_count(self):
        return self["query-count"]

    @property
    def duplicate_count(self):
        return self["duplicates"]

    def _get_key(self, key, default):
        return self._raw_data.get(key, default)

    def _get_required_key(self, key):
        if key in self._raw_data:
            return self._raw_data.get(key)
        raise_error("Got invalid data. It is missing a required key: %s" % key)


def iter_entries(entries):
    for module_name, module_data in sorted(entries.items()):
        assert_type(module_data, dict)

        yield module_name, (
            Entry(test_name, module_name, test_data)
            for test_name, test_data in sorted(module_data.items())
        )


def flatten_entries(file_content):
    entries = []
    for _, data in iter_entries(file_content):
        entries += list(data)
    return entries



# coding=utf-8
from collections import namedtuple

from pytest_django_queries.entry import Entry  # noqa
from pytest_django_queries.filters import format_underscore_name_to_human

_ROW_FIELD = namedtuple("_RowField", ("comp_field", "align_char", "length_field"))
_ROW_FIELDS = (
    _ROW_FIELD("test_name", "<", "test_name"),
    _ROW_FIELD("left_count", ">", "query_count"),
    _ROW_FIELD("right_count", ">", "query_count"),
    _ROW_FIELD("duplicate_count", ">", "duplicate_count"),
)
_ROW_PREFIX = "  "
_NA_CHAR = "-"


def entry_row(entry_comp, lengths):
    cols = []

    for field, align, length_key in _ROW_FIELDS:
        fmt = "{cmp.%s: %s%d}" % (field, align, lengths[length_key])
        cols.append(fmt.format(cmp=entry_comp, lengths=lengths))

    return "%(diff_char)s %(results)s" % (
        {"diff_char": entry_comp.diff, "results": "\t".join(cols)}
    )


def get_header_row(lengths):
    sep_row = []
    head_row = []

    for field, _, length_key in _ROW_FIELDS:
        length = lengths[length_key]
        sep_row.append("%s" % ("-" * length))
        head_row.append(
            "{field: <{length}}".format(field=field.replace("_", " "), length=length)
        )

    return "%(prefix)s%(head)s\n%(prefix)s%(sep)s" % (
        {"prefix": _ROW_PREFIX, "head": "\t".join(head_row), "sep": "\t".join(sep_row)}
    )


class DiffChars(object):
    NEGATIVE = "-"
    NEUTRAL = " "
    POSITIVE = "+"

    @classmethod
    def convert(cls, diff):
        if diff < 0:
            return DiffChars.POSITIVE
        if diff > 0:
            return DiffChars.NEGATIVE
        return DiffChars.NEUTRAL


class SingleEntryComparison(object):
    __slots__ = ["left", "right"]

    def __init__(self, left=None, right=None):
        """
        :param left: Previous version.
        :type left: Entry

        :param right: Newest version.
        :type right: Entry
        """

        self.left = left
        self.right = right

    def _diff_from_newest(self):
        """
        Returns the query count difference from the previous version.
        If there is no older version, we assume it's an "improvement" (positive output)
        If there is no new version, we assume it's not an improvement (negative output)
        """
        if self.left is None:
            return DiffChars.POSITIVE
        if self.right is None:
            return DiffChars.NEGATIVE
        return DiffChars.convert(self.right.query_count - self.left.query_count)

    @property
    def test(self):
        return self.left or self.right

    @property
    def test_name(self):
        return format_underscore_name_to_human(self.test.test_name)

    @property
    def left_count(self):
        return str(self.left.query_count) if self.left else _NA_CHAR

    @property
    def right_count(self):
        return str(self.right.query_count) if self.right else _NA_CHAR

    @property
    def duplicate_count(self):
        if self.right:
            return self.right.duplicate_count
        elif self.left:
            return self.left.duplicate_count
        return _NA_CHAR

    @property
    def diff(self):
        return self._diff_from_newest()

    def to_string(self, lengths):
        return entry_row(self, lengths=lengths)


class DiffGenerator(object):
    def __init__(self, entries_left, entries_right):
        """
        Generates the diffs from two files.

        :param entries_left:
        :type entries_left: List[Entry]
        :param entries_right:
        :type entries_right: List[Entry]
        """

        self.entries_left = entries_left
        self.entries_right = entries_right

        self._mapping = {}
        self._generate_mapping()
        self.longest_props = self._get_longest_per_prop()
        self.header_rows = get_header_row(lengths=self.longest_props)

    def _get_longest_per_prop(self):
        longest = {field.length_field: len(field.comp_field) for field in _ROW_FIELDS}
        entries = self.entries_left + self.entries_right

        for entry in entries:
            for field in _ROW_FIELDS:
                current_length = len(str(getattr(entry, field.comp_field, None)))
                if current_length > longest[field.length_field]:
                    longest[field.length_field] = current_length

        return longest

    def _map_side(self, entries, side_name):
        for entry in entries:
            module_map = self._mapping.setdefault(entry.module_name, {})

            if entry.test_name not in module_map:
                module_map[entry.test_name] = SingleEntryComparison()

            setattr(module_map[entry.test_name], side_name, entry)

    def _generate_mapping(self):
        self._map_side(self.entries_left, "left")
        self._map_side(self.entries_right, "right")

    def _iter_module(self, module_entries):
        yield self.header_rows
        for _, test_comparison in sorted(
            module_entries.items()
        ):  # type: SingleEntryComparison
            yield test_comparison.to_string(lengths=self.longest_props)

    def _iter_modules(self):
        for module_name, module_entries in sorted(self._mapping.items()):
            yield (
                format_underscore_name_to_human(module_name),
                self._iter_module(module_entries),
            )

    def __iter__(self):
        return self._iter_modules()



import shutil
import sys

import click


def raise_error(message, exit_code=1):
    click.echo("Error: %s" % message, file=sys.stderr)
    sys.exit(exit_code)


def assert_type(value, expected_type):
    if type(value) != expected_type:
        raise_error(
            "Expected a %s, got %s instead"
            % (expected_type.__name__, type(value).__name__)
        )


def create_backup(save_path, backup_path):
    shutil.copy(save_path, backup_path)



import click
from beautifultable import BeautifulTable

from pytest_django_queries.entry import Entry, iter_entries
from pytest_django_queries.filters import format_underscore_name_to_human


def print_entries(data):
    table = BeautifulTable()
    table.column_headers = ["Module", "Tests"]
    for module_name, module_entries in iter_entries(data):
        subtable = BeautifulTable()
        subtable.column_headers = [field for _, field in Entry.FIELDS]
        for entry in module_entries:
            subtable.append_row([getattr(entry, field) for field, _ in Entry.FIELDS])
        table.append_row([module_name, subtable])
    click.echo(table)


def entries_to_html(data, template):
    html_content = template.render(
        data=iter_entries(data), humanize=format_underscore_name_to_human
    )
    return html_content



import json
from os.path import abspath, dirname
from os.path import join as pathjoin

import click
from jinja2 import Template
from jinja2 import exceptions as jinja_exceptions

from pytest_django_queries.diff import DiffGenerator
from pytest_django_queries.entry import flatten_entries
from pytest_django_queries.plugin import (
    DEFAULT_OLD_RESULT_FILENAME,
    DEFAULT_RESULT_FILENAME,
)
from pytest_django_queries.tables import entries_to_html, print_entries
from pytest_django_queries.utils import create_backup

HERE = dirname(__file__)
DEFAULT_TEMPLATE_PATH = abspath(pathjoin(HERE, "templates", "default_bootstrap.jinja2"))
DEFAULT_HTML_SAVE_PATH = "django-queries-results.html"

DIFF_TERM_COLOR = {"-": "red", "+": "green"}
DEFAULT_TERM_DIFF_COLOR = None


def _write_html_to_file(content, path):
    with open(path, "w") as fp:
        fp.write(content)


class JsonReportFileParamType(click.File):
    name = "report_file"

    def convert(self, value, param, ctx):
        fp = super(JsonReportFileParamType, self).convert(value, param, ctx)
        try:
            loaded = json.load(fp)
            if type(loaded) is not dict:
                self.fail("The file is not a dictionary", param, ctx)
            return loaded
        except ValueError as e:
            self.fail("The file is not valid json: %s" % str(e), param, ctx)


class Jinja2TemplateFile(click.File):
    name = "jinja2_file"

    def convert(self, value, param, ctx):
        fp = super(Jinja2TemplateFile, self).convert(value, param, ctx)
        try:
            return Template(fp.read(), trim_blocks=True)
        except jinja_exceptions.TemplateError as e:
            self.fail(
                "The file is not a valid jinja2 template: %s" % str(e), param, ctx
            )


@click.group()
def main():
    """Command line tool for pytest-django-queries."""


@main.command()
@click.argument(
    "input_file", type=JsonReportFileParamType("r"), default=DEFAULT_RESULT_FILENAME
)
def show(input_file):
    """View a given rapport."""
    return print_entries(input_file)


@main.command()
@click.argument(
    "input_file", type=JsonReportFileParamType("r"), default=DEFAULT_RESULT_FILENAME
)
@click.option(
    "-o",
    "--output",
    type=str,
    default=DEFAULT_HTML_SAVE_PATH,
    help="The path to save the HTML file into django-queries.html by default."
    "You can pass a dash (-) to write to stdout as well.",
)
@click.option(
    "--template",
    type=Jinja2TemplateFile("r"),
    default=DEFAULT_TEMPLATE_PATH,
    help="Use a custom jinja2 template for rendering HTML results.",
)
def html(input_file, output, template):
    """
    Render the results as HTML instead of a raw table.

    Note: you can pass a dash (-) as the path to print the HTML content to stdout."""
    html_content = entries_to_html(input_file, template)

    if output == "-":
        click.echo(html_content, nl=False)
        return

    _write_html_to_file(html_content, output)


@main.command()
@click.argument(
    "left_file", type=JsonReportFileParamType("r"), default=DEFAULT_OLD_RESULT_FILENAME
)
@click.argument(
    "right_file", type=JsonReportFileParamType("r"), default=DEFAULT_RESULT_FILENAME
)
def diff(left_file, right_file):
    """Render the diff as a console table with colors."""
    left = flatten_entries(left_file)
    right = flatten_entries(right_file)
    first_line = True
    for module_name, lines in DiffGenerator(left, right):
        if not first_line:
            click.echo()
        else:
            first_line = False

        click.echo("# %s" % module_name)
        for line in lines:
            fg_color = DIFF_TERM_COLOR.get(line[0], DEFAULT_TERM_DIFF_COLOR)
            click.secho(line, fg=fg_color)


@main.command()
@click.argument("target_path", type=str, default=DEFAULT_OLD_RESULT_FILENAME)
def backup(target_path):
    source_path = DEFAULT_RESULT_FILENAME
    click.echo("{0} -> {1}".format(source_path, target_path))
    create_backup(source_path, target_path)


if __name__ == "__main__":
    main()



import json
import os.path
import shutil
import tempfile
from os import listdir
from os.path import isfile

import pytest
from django.test.utils import CaptureQueriesContext

# Defines the plugin marker name
from pytest_django_queries.utils import create_backup

PYTEST_QUERY_COUNT_MARKER = "count_queries"
PYTEST_QUERY_COUNT_FIXTURE_NAME = "count_queries"
DEFAULT_RESULT_FILENAME = ".pytest-queries"
DEFAULT_OLD_RESULT_FILENAME = ".pytest-queries.old"


def is_slave(config):
    return hasattr(config, "slaveinput")


def get_slaveid(config):
    if hasattr(config, "slaveinput"):
        return config.workerinput["slaveid"]
    else:
        return "master"


def save_results_to_json(save_path, backup_path, data):
    if backup_path and isfile(save_path):
        create_backup(save_path, backup_path)

    with open(save_path, "w") as fp:
        json.dump(data, fp, indent=2)


def add_entry(request, queries, dirout):
    module_name = request.node.module.__name__
    test_name = request.node.name
    queries = queries[:]
    query_count = len(queries)
    duplicate_count = query_count - len(set((q["sql"] for q in queries)))

    result_line = "%s\t%s\t%d\t%d\n" % (
        module_name,
        test_name,
        query_count,
        duplicate_count,
    )
    save_path = os.path.join(dirout, get_slaveid(request.config))
    if os.path.isfile(save_path):
        mode = "a"
    else:
        mode = "w"

    with open(save_path, mode=mode) as fp:
        fp.write(result_line)


def pytest_addoption(parser):
    group = parser.getgroup("django-queries")
    group.addoption(
        "--django-db-bench",
        dest="queries_results_save_path",
        action="store",
        default=DEFAULT_RESULT_FILENAME,
        metavar="PATH",
        help="Output file for storing the results. Default: .pytest-queries",
    )
    group.addoption(
        "--django-backup-queries",
        dest="queries_backup_results",
        action="store",
        default=None,
        metavar="PATH",
        help="Whether the old results should be backed up or not before overriding",
        nargs="?",
    )


@pytest.mark.tryfirst
def pytest_load_initial_conftests(early_config, parser, args):
    """
    :param early_config:
    :param parser:
    :param args:
    :type args: tuple|list
    :return:
    """
    early_config.addinivalue_line(
        "markers",
        "%s: Mark the test as to have their database query counted."
        "" % PYTEST_QUERY_COUNT_MARKER,
    )

    backup_path = early_config.known_args_namespace.queries_backup_results

    # Set default value if the flag was provided without value in arguments
    if backup_path is None and "--django-backup-queries" in args:
        backup_path = DEFAULT_OLD_RESULT_FILENAME

    early_config.known_args_namespace.queries_backup_results = backup_path


def _process_query_count_marker(request, *_args, **kwargs):
    autouse = kwargs.setdefault("autouse", True)
    if autouse:
        request.getfixturevalue(PYTEST_QUERY_COUNT_FIXTURE_NAME)


@pytest.fixture(autouse=True)
def _pytest_query_marker(request):
    """Use the fixture to count the queries on the current node if it's
    marked with 'count_queries'.

    Optional keyword-arguments:
        - autouse (bool, default: True)
          Whether the fixture should be used automatically.
          This might be useful if you are executing fixtures
          that are making queries and still want to mark the test
          but place the fixture manually."""
    marker = request.node.get_closest_marker(PYTEST_QUERY_COUNT_MARKER)
    if marker:
        _process_query_count_marker(request, *marker.args, **marker.kwargs)


def pytest_configure(config):
    config.django_queries_shared_directory = tempfile.mkdtemp(
        prefix="pytest-django-queries"
    )


def pytest_unconfigure(config):
    results_path = config.django_queries_shared_directory
    test_results = {}

    for filename in listdir(results_path):
        with open(os.path.join(results_path, filename)) as fp:
            for result_line in fp.readlines():
                result_line = result_line.strip()

                if not result_line:
                    continue

                module_name, test_name, query_count, duplicates = result_line.split(
                    "\t"
                )

                module_entries = test_results.setdefault(module_name, {})
                module_entries[test_name] = {
                    "query-count": int(query_count),
                    "duplicates": int(duplicates),
                }

    if test_results:
        save_results_to_json(
            save_path=config.known_args_namespace.queries_results_save_path,
            backup_path=config.known_args_namespace.queries_backup_results,
            data=test_results,
        )

    # clean up the temporary directory
    shutil.rmtree(config.django_queries_shared_directory)


def pytest_configure_node(node):
    node.slaveinput[
        "_django_queries_shared_dir"
    ] = node.config.django_queries_shared_directory


pytest_configure_node.optionalhook = True


def get_shared_directory(request):
    """Returns a unique and temporary directory which can be shared by
    master or worker nodes in xdist runs.
    """
    if not is_slave(request.config):
        return request.config.django_queries_shared_directory
    else:
        return request.config.slaveinput["_django_queries_shared_dir"]


@pytest.fixture
def count_queries(request):
    """Wrap a test to count the number of performed queries."""
    from django.db import connection

    with CaptureQueriesContext(connection) as context:
        yield context
    add_entry(request, context, get_shared_directory(request))
